@@ -47,6 +47,9 @@
 // ------------------------------------------------------------------------
 
 bool g_verbose = false;
+bool g_skip_identical = false;
+bool g_mark_differences = false;
+long g_channel_tolerance = 0;
 
 // Resolution to use for rasterization, in DPI
 #define RESOLUTION  300
@@ -179,6 +182,8 @@ cairo_surface_t *diff_images(cairo_surface_t *s1, cairo_surface_t *s2,
               y < r2.height;
               y++, data2 += stride2, out += stridediff )
         {
+            bool linediff = false;
+
             for ( int x = 0; x < r2.width * 4; x += 4 )
             {
                 unsigned char cr1 = *(out + x + 0);
@@ -189,9 +194,15 @@ cairo_surface_t *diff_images(cairo_surface_t *s1, cairo_surface_t *s2,
                 unsigned char cg2 = *(data2 + x + 1);
                 unsigned char cb2 = *(data2 + x + 2);
 
-                if ( cr1 != cr2 || cg1 != cg2 || cb1 != cb2 )
+
+		if ( cr1 > (cr2+g_channel_tolerance) || cr1 < (cr2-g_channel_tolerance)
+		  || cg1 > (cg2+g_channel_tolerance) || cg1 < (cg2-g_channel_tolerance)
+		  || cb1 > (cb2+g_channel_tolerance) || cb1 < (cb2-g_channel_tolerance)
+		   )
                 {
                     changes = true;
+                    linediff = true;
+
                     if ( thumbnail )
                     {
                         // calculate the coordinates in the thumbnail
@@ -212,6 +223,14 @@ cairo_surface_t *diff_images(cairo_surface_t *s1, cairo_surface_t *s2,
                 // change the B channel to be from s2; RG will be s1
                 *(out + x + 2) = cb2;
             }
+
+            if (g_mark_differences && linediff) {
+                for (int x = 0; x < (10 < r2.width ? 10 : r2.width) * 4; x+=4) {
+                   *(out + x + 0) = 0;
+                   *(out + x + 1) = 0;
+                   *(out + x + 2) = 255;
+		}
+            }
         }
     }
 
@@ -322,10 +341,12 @@ bool page_compare(cairo_t *cr_out,
             // by writing unchanged pages in their original form rather than
             // a rasterized one
 
-            poppler_page_render(page1, cr_out);
+            if (!g_skip_identical)
+               poppler_page_render(page1, cr_out);
         }
 
-        cairo_show_page(cr_out);
+        if (diff || !g_skip_identical)
+            cairo_show_page(cr_out);
     }
 
     if ( diff )
@@ -351,7 +372,7 @@ bool doc_compare(PopplerDocument *doc1, PopplerDocument *doc2,
                  wxProgressDialog *progress = NULL,
                  Gutter *gutter = NULL)
 {
-    bool are_same = true;
+    int pages_differ = 0;
 
     cairo_surface_t *surface_out = NULL;
     cairo_t *cr_out = NULL;
@@ -372,7 +393,6 @@ bool doc_compare(PopplerDocument *doc1, PopplerDocument *doc2,
     {
         if ( g_verbose )
             printf("pages count differs: %d vs %d\n", pages1, pages2);
-        are_same = false;
     }
 
     for ( int page = 0; page < pages_total; page++ )
@@ -450,7 +470,7 @@ bool doc_compare(PopplerDocument *doc1, PopplerDocument *doc2,
 
         if ( !page_same )
         {
-            are_same = false;
+	    pages_differ ++;
 
             if ( g_verbose )
                 printf("page %d differs\n", page);
@@ -470,7 +490,11 @@ bool doc_compare(PopplerDocument *doc1, PopplerDocument *doc2,
         cairo_surface_destroy(surface_out);
     }
 
-    return are_same;
+    if (g_verbose)
+        printf("%d of %d pages differ.\n", pages_differ, pages_total);
+
+    // are doc1 and doc1 the same?
+    return (pages_differ == 0) && (pages1 == pages2);
 }
 
 
@@ -859,10 +883,20 @@ int main(int argc, char *argv[])
         { wxCMD_LINE_SWITCH,
                   wxT28("v"), wxT28("verbose"), wxT28("be verbose") },
 
+        { wxCMD_LINE_SWITCH,
+                  wxT28("s"), wxT28("skip-identical"), wxT28("only output pages with differences") },
+
+        { wxCMD_LINE_SWITCH,
+                  wxT28("m"), wxT28("mark-differences"), wxT28("additionally mark differences on left side") },
+
         { wxCMD_LINE_OPTION,
                   NULL, wxT28("output-diff"), wxT28("output differences to given PDF file"),
                   wxCMD_LINE_VAL_STRING },
 
+        { wxCMD_LINE_OPTION,
+                  NULL, wxT28("channel-tolerance"), wxT28("consider channel values to be equal if within specified tolerance"),
+                  wxCMD_LINE_VAL_NUMBER },
+
         { wxCMD_LINE_SWITCH,
                   NULL, wxT28("view"), wxT28("view the differences in a window") },
 
@@ -894,6 +928,12 @@ int main(int argc, char *argv[])
     if ( parser.Found(wxT("verbose")) )
         g_verbose = true;
 
+    if ( parser.Found(wxT("skip-identical")) )
+        g_skip_identical = true;
+
+    if ( parser.Found(wxT("mark-differences")) )
+        g_mark_differences = true;
+
     wxFileName file1(parser.GetParam(0));
     wxFileName file2(parser.GetParam(1));
     file1.MakeAbsolute();
@@ -919,6 +959,15 @@ int main(int argc, char *argv[])
         return 3;
     }
 
+    if ( parser.Found(wxT("channel-tolerance"), &g_channel_tolerance) )
+    {
+        if (g_channel_tolerance < 0 || g_channel_tolerance > 255) {
+            fprintf(stderr, "Invalid channel-tolerance: %ld. Valid range is 0(default, exact matching)-255\n", g_channel_tolerance);
+            return 2;
+	}
+    }
+
+
     int retval = 0;
 
     wxString pdf_file;
